-----------------------------------------------------------------------------------------------
-- Client Lua Script for BossMods
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
require "Window"
require "GameLib"
require "Apollo"
-----------------------------------------------------------------------------------------------
-- BossMods Module Definition
-----------------------------------------------------------------------------------------------
local BossMods = {} 

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local nLastTime, nRefreshTime, fRefresh = 0, 0, 0.016
local healthPercent = 100
local bTimer = false
local fTimerStart = {0, 0}
local bossUnit = nil
local nGlobalCounter = 1
local tEncounter = {}
local tBoss = {}
local nCurrentPhase = 1
local bReset = false
local BM_settings = {{nil, nil}, true}
local bMovable = false
local spellName = nil
local lastSent = nil
local fVersion = 0.37
local bmLine = true
local bmRear = true
local bmRalways = false
local bmReply = true
local bmNF = false
local bmOD = false


-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function BossMods:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
    return o
end

function BossMods:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

-----------------------------------------------------------------------------------------------
-- BossMods OnLoad
-----------------------------------------------------------------------------------------------
function BossMods:OnLoad()
	self.xmlDoc = XmlDoc.CreateFromFile("BossMods.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- BossMods OnDocLoaded
-----------------------------------------------------------------------------------------------

function BossMods:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		self.BossOptions = Apollo.LoadForm(self.xmlDoc, "BossOptions", nil, self)
		self.BarContainer = Apollo.LoadForm(self.xmlDoc, "BarContainer", nil, self)
		self.Background = Apollo.LoadForm(self.xmlDoc, "Background", "InWorldHudStratum", self)
		self.BMPhase = self.BarContainer:FindChild("BMPhase")
		self.BMTimer = {{false, self.BarContainer:FindChild("BMTimer"), nil, nil}, {false, self.BarContainer:FindChild("BMTimer2"), nil, nil}, {false, self.BarContainer:FindChild("BMTimer3"), nil, nil}}
		if self.BarContainer == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
	    self.BarContainer:Show(true)
		self.BossOptions:Show(false, true)
		self.BMPhase:Show(false, true)
		self.BMTimer[1][2]:Show(false, true)
		self.BMTimer[2][2]:Show(false, true)
		self.BMTimer[3][2]:Show(false, true)
		
		if BM_settings[1][1] then
			self.BarContainer:SetAnchorOffsets(unpack(BM_settings[1][1]))
		end
		if BM_settings[1][2] then
			self.BossOptions:SetAnchorOffsets(unpack(BM_settings[1][2]))
		end
		-- Register handlers for events, slash commands and timer, etc.
		Apollo.RegisterSlashCommand("boss", "OnBossModsOn", self)
		Apollo.RegisterSlashCommand("bossline", "OnBossLineOn", self)
		Apollo.RegisterSlashCommand("bossrear", "OnBossRearOn", self)
		Apollo.RegisterSlashCommand("bossreset", "OnBossModsOn", self)
		Apollo.RegisterEventHandler("ChatMessage","OnChatMessage", self)
		Apollo.RegisterEventHandler("TargetUnitChanged", "OnTargetUnit", self)
		tBoss = self.BMLocale:Fights()
		if bmRalways then
			Apollo.RegisterEventHandler("NextFrame", "OnUpdate", self)
			bmNF = true
		end
	end
end

-----------------------------------------------------------------------------------------------
-- BossMods Functions
-----------------------------------------------------------------------------------------------
-- on SlashCommand "/bm"
function BossMods:OnBossModsOn(msg)
	if msg == "bossreset" then
		bReset = true
		RequestReloadUI()
	else
		if bmLine then
			self.BossOptions:FindChild("Lines"):SetCheck(true)
		end
		if bmRear then
			self.BossOptions:FindChild("Rear"):SetCheck(true)
		end
		if bmReply then
			self.BossOptions:FindChild("Reply"):SetCheck(true)
		end
		if bmRalways then
			self.BossOptions:FindChild("Target"):SetCheck(true)
		end
		self.BossOptions:FindChild("Title"):SetText("BossMods v"..fVersion)
		self.BossOptions:Show(true)
	end
end

function BossMods:OnSave(eLevel)
    if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return nil
    end
	if not bReset then
		BM_settings = {{{self.BarContainer:GetAnchorOffsets()}, {self.BossOptions:GetAnchorOffsets()}}, bmLine, bmRear, bmReply, bmRalways}
		BM_settings[0] = fVersion
	    return BM_settings
	else
		return {reset = true}
	end
end

function BossMods:OnRestore(eLevel, tData)
	if tData then
		if not tData["reset"] then
			if tData[1] then
				BM_settings = tData
				bmLine = tData[2]
				bmRear = tData[3]
				bmReply = tData[4]
				if tData[5] then
					bmRalways = tData[5]
				end
			end
			if not tData[0] then
				bmLine = true
				bmRear = true
				bmReply = true
				bmRalways = false
			end
		end
	end
end

function BossMods:OnTargetUnit(unitTarget)
	if unitTarget then
		for i=1, #tBoss do
			if unitTarget:GetName() == tBoss[i][2] then
				if not tBoss[i][1] then
					if not bmNF then
						Apollo.RegisterEventHandler("NextFrame", "OnUpdate", self)
						bmNF = true
					end
					if not bmOD then
						Apollo.RegisterEventHandler("CombatLogDamage", "OnDamage", self)
						bmOD = true
					end
					tBoss[i][1] = true
					bossUnit = unitTarget
					tEncounter = tBoss[i]
				else
					break
				end
			end
		end
		if not bmNF and bmRalways then
			Apollo.RegisterEventHandler("NextFrame", "OnUpdate", self)
			bmNF = true
		end
	end
end

function BossMods:OnDamage(tMessage)
	if tEncounter then
		if tEncounter[5] then
			for i=1, #tEncounter[5] do
				if tMessage.splCallingSpell:GetName()== tEncounter[5][i][2] then
					if tEncounter[5][i][1] then
						for i1=1, #self.BMTimer do
							if self.BMTimer[i1][3] == 5 and self.BMTimer[i1][4] == i then
								local fCheckTime = fTimerStart[i1] - 5
								if  fCheckTime < GameLib.GetGameTime() then
									self:CreateBar(tEncounter[5][i], 5, i)
								end
								break
							end
						end
					else
						self:CreateBar(tEncounter[5][i], 5, i)
					end
					break
				end
			end
		end
	end
end

function BossMods:CreateBar(tBar, nType, nEvent)
	local bInPhase = false
	local nTimer = 0
	for i=1, #tBar[5] do 
		if nCurrentPhase == tBar[5][i] then
			bInPhase = true
			break
		end
	end
	if bInPhase then
		if tBar[4] then
				nTimer = tBar[4]
		end
		if tBar[7] then
			if tBar[11] < tBar[9] then
				tEncounter[nType][nEvent][11] = tEncounter[nType][nEvent][11] + 1
			else
				tEncounter[nType][nEvent][11] = 1
				nTimer = tBar[10]
			end
		end
		if tBar[8] then
			if nGlobalCounter < tBar[9] then
				nGlobalCounter = nGlobalCounter + 1
			else
				nGlobalCounter = 1
				nTimer = tBar[10]
			end
		end
		local whichBar = 1
		if tBar[1] then
			whichBar = tBar[6]
		else
			for i=1, #self.BMTimer do 
			 	if not self.BMTimer[i][1] then
					whichBar = i
					break
				end
			end
		end
		self.BMTimer[whichBar][3] = nType
		self.BMTimer[whichBar][4] = nEvent
		self.BMTimer[whichBar][1] = true
		tEncounter[nType][nEvent][1] = true
		tEncounter[nType][nEvent][6] = whichBar
		self.BMTimer[whichBar][2]:SetMax(nTimer)
		self.BMTimer[whichBar][2]:SetText(tBar[3])
		self.BMTimer[whichBar][2]:SetProgress(nTimer)
		self.BMTimer[whichBar][2]:Invoke()
		fTimerStart[whichBar] = GameLib.GetGameTime() + nTimer
	end
end

function BossMods:OnChatMessage(channelCurrent, tMessage)
	if channelCurrent:GetName() == "NPC Say" then
		for i=1, #tBoss do
			if tMessage.strSender == tBoss[i][2] then
				if not tBoss[i][1] and tMessage.unitSource:IsInCombat() then
					if not bmNF then
						Apollo.RegisterEventHandler("NextFrame", "OnUpdate", self)
						bmNF = true
					end
					if not bmOD then
						Apollo.RegisterEventHandler("CombatLogDamage", "OnDamage", self)
						bmOD = true
					end
					tBoss[i][1] = true
					bossUnit = tMessage.unitSource
					tEncounter = tBoss[i]
				end
				break
			end
		end
	elseif channelCurrent:GetName() == "Datachron" then
		if tEncounter then
			if tEncounter[6] then
				for i=1, #tEncounter[6]do
					if string.find(tMessage.arMessageSegments[1].strText, tEncounter[6][i][2]) then
						self:CreateBar(tEncounter[6][i], 6, i)
						break
					end
				end
			end
		end
	elseif channelCurrent:GetName() == "Whisper" and bmReply then
		if tEncounter then
			if bossUnit then
				if lastSent ~= tMessage.strSender and not tMessage.bSelf then
					lastSent = tMessage.strSender
					ChatSystemLib.Command("/w "..tMessage.strSender.." Boss Mods Auto Reply: Currently Fighting '"..tEncounter[2].."' at "..math.ceil(healthPercent).."% Health.")
				end
			end
		end
	end
end

function BossMods:OnUpdate()
	local nCurrentTime = GameLib.GetGameTime()
	local nElapsedTime = nCurrentTime - nLastTime
	nLastTime = nCurrentTime
	nRefreshTime = nRefreshTime + nElapsedTime
	if nRefreshTime >= fRefresh then 
		if bmRalways then
			self.Background:DestroyAllPixies()
			local myTarget = GameLib.GetTargetUnit()
			if myTarget then
				if myTarget:IsInCombat() and myTarget:GetDispositionTo(GameLib.GetPlayerUnit()) < 2 then
					local nAngle = myTarget:GetHeading()
					local gPos = myTarget:GetPosition()
					local vBoss = Vector3.New(gPos.x, gPos.y, gPos.z)
					local bossPos = GameLib.WorldLocToScreenPoint(vBoss)
					local tRadians = {-1.9, 1.9}
					for i=1, 2 do
						local fAngle = nAngle+tRadians[i]
						local newX = math.sin(fAngle) * 10
						local newZ = math.cos(fAngle) * 10
						local vLine = Vector3.New(gPos.x-newX, gPos.y, gPos.z-newZ)
						local bmPos = GameLib.WorldLocToScreenPoint(vLine)
						self.Background:AddPixie( {
							bLine = true, fWidth = 4, cr = "blue",
							loc = { fPoints = { 0, 0, 0, 0 }, nOffsets = { bmPos.x, bmPos.y, bossPos.x, bossPos.y} }
						} )
					end
				end
			end
		end
		if tEncounter then
			if bossUnit then
				if bmLine or bmRear then
					if not bmRalways then
						self.Background:DestroyAllPixies()
					end
					local nAngle = bossUnit:GetHeading()
					local gPos = bossUnit:GetPosition()
					if gPos and nAngle then
						if tEncounter[7] and bmLine then
							for i=1, #tEncounter[7] do
								local bossX = math.sin(nAngle+tEncounter[7][i][4]) * tEncounter[7][i][3]
								local bossZ = math.cos(nAngle+tEncounter[7][i][4]) * tEncounter[7][i][3]
								local vBoss = Vector3.New(gPos.x-bossX, gPos.y, gPos.z-bossZ)
								local bossPos = GameLib.WorldLocToScreenPoint(vBoss)
								local fAngle = nAngle+tEncounter[7][i][1]
								local newX = math.sin(fAngle) * 15
								local newZ = math.cos(fAngle) * 15
								local vLine = Vector3.New(gPos.x-newX-bossX, gPos.y, gPos.z-newZ-bossZ)
								local bmPos = GameLib.WorldLocToScreenPoint(vLine)
								self.Background:AddPixie( {
									bLine = true, fWidth = 4, cr = tEncounter[7][i][2],
									loc = { fPoints = { 0, 0, 0, 0 }, nOffsets = { bmPos.x, bmPos.y, bossPos.x, bossPos.y} }
								} )
							end
						end
						if bmRear then
							if bmRalways and bossUnit == GameLib.GetTargetUnit then
								-- Do Nothing, lines already drawn
							else
								local vBoss = Vector3.New(gPos.x, gPos.y, gPos.z)
								local bossPos = GameLib.WorldLocToScreenPoint(vBoss)
								local tRadians = {-1.9, 1.9}
								for i=1, 2 do
									local fAngle = nAngle+tRadians[i]
									local newX = math.sin(fAngle) * 10
									local newZ = math.cos(fAngle) * 10
									local vLine = Vector3.New(gPos.x-newX, gPos.y, gPos.z-newZ)
									local bmPos = GameLib.WorldLocToScreenPoint(vLine)
									self.Background:AddPixie( {
										bLine = true, fWidth = 4, cr = "blue",
										loc = { fPoints = { 0, 0, 0, 0 }, nOffsets = { bmPos.x, bmPos.y, bossPos.x, bossPos.y} }
									} )
								end
							end
						end
					end
				end

				local bossSpell = bossUnit:GetCastName()
				local bossCasting = bossUnit:IsCasting()
				
				if not bossCasting then
					spellName = nil
				end
				
				if spellName == nil then
					for i=1, #tEncounter[4] do
						if tEncounter[4][i][2] == bossSpell then
							--Print(tEncounter[4][i][2].." - Spell Hit 1")
							bInPhase = false
							for i1=1, #tEncounter[4][i][5] do 
								if nCurrentPhase == tEncounter[4][i][5][i1] then
									bInPhase = true
									break
								end
							end
							if bInPhase then
								--Print(tEncounter[4][i][2].." - Spell Hit 2")
								self:CreateBar(tEncounter[4][i], 4, i)
								spellName = bossSpell
							end
						end
					end
				end
				local playerUnit = GameLib.GetPlayerUnit()
				if tEncounter[3][1] and not bossUnit:IsDead() and bossUnit and bossUnit:GetHealth() then
					healthPercent = (100 / bossUnit:GetMaxHealth()) * bossUnit:GetHealth()
					for i=1, #tEncounter[3][2] do
						if healthPercent > tEncounter[3][2][i] then
							nCurrentPhase = i
							local nMax = 100-tEncounter[3][2][i]
							if i ~= 1 then
								nMax = tEncounter[3][2][i-1] - tEncounter[3][2][i]
							end
							if tEncounter[3][3] then
								self.BMPhase:SetText(tEncounter[3][4][i])
							else
								self.BMPhase:SetText("Phase: "..i.."/"..#tEncounter[3][2])
							end
							self.BMPhase:SetMax(nMax)
							
							local nNewNum = nMax - (100 - healthPercent)
							if i ~= 1 then
								nNewNum = nMax - (tEncounter[3][2][i-1] - healthPercent)
							end
							self.BMPhase:SetProgress(nNewNum)
							self.BMPhase:Invoke()
							break
						else
							self.BMPhase:Show(false)
						end
					end
				else
					for i=1, #tBoss do
						if tBoss[i][2] == tEncounter[2] then
							tBoss[i][1] = false
							break
						end
					end
					tEncounter = nil
					self.BMPhase:Show(false)
					if not bmRalways then
						Apollo.RemoveEventHandler("NextFrame", self)
						bmNF = false
					end
					Apollo.RemoveEventHandler("CombatLogDamage", self)
					bmOD = false
					lastSent = nil
					self.Background:DestroyAllPixies()
					for i=1, #self.BMTimer do
						self.BMTimer[i][1] = false
						self.BMTimer[i][2]:Show(false)
					end
				end
				for i=1, #self.BMTimer do
					if self.BMTimer[i][1] then
						if nCurrentTime <= fTimerStart[i] then
							local progresstime = fTimerStart[i] - nCurrentTime
							self.BMTimer[i][2]:SetProgress(progresstime)
							local myText = tEncounter[self.BMTimer[i][3]][self.BMTimer[i][4]][3]
							self.BMTimer[i][2]:SetText(myText.." ("..math.ceil(progresstime)..")")
						else
							self.BMTimer[i][1] = false
							self.BMTimer[i][2]:Show(false)
							tEncounter[self.BMTimer[i][3]][self.BMTimer[i][4]][1] = false
						end
					end
				end
			end	
		end
		nRefreshTime = 0
	end
end

---------------------------------------------------------------------------------------------------
-- BossOptions Functions
---------------------------------------------------------------------------------------------------

function BossMods:OnLine( wndHandler, wndControl, eMouseButton )
	if wndHandler:IsChecked() then
		bmLine = true
	else
		bmLine = false
		self.Background:DestroyAllPixies()
	end
end

function BossMods:OnRear( wndHandler, wndControl, eMouseButton )
	if wndHandler:IsChecked() then
		bmRear = true
	else
		bmRear = false
		self.Background:DestroyAllPixies()
	end
end

function BossMods:OnBar( wndHandler, wndControl, eMouseButton )
	if wndHandler:IsChecked() then
		self.BMPhase:Show(true)
		self.BMTimer[1][2]:Show(true)
		self.BMTimer[2][2]:Show(true)
		self.BMTimer[3][2]:Show(true)
		bMovable = true
		self.BarContainer:SetStyle("Moveable", true)
		self.BarContainer:SetStyle("Picture", true)
		self.BarContainer:SetStyle("IgnoreMouse", false)
		self.BarContainer:SetText("BossMods v"..fVersion)
	else
		bMovable = false
		self.BarContainer:SetStyle("Moveable", false)
		self.BarContainer:SetStyle("Picture", false)
		self.BarContainer:SetStyle("IgnoreMouse", true)
		self.BarContainer:SetText("")
		self.BMPhase:Show(false)
		self.BMTimer[1][2]:Show(false)
		self.BMTimer[2][2]:Show(false)
		self.BMTimer[3][2]:Show(false)
	end
end

function BossMods:OnReply( wndHandler, wndControl, eMouseButton )
	if wndHandler:IsChecked() then
		bmReply = true
	else
		bmReply = false
	end
end

function BossMods:OnClose( wndHandler, wndControl, eMouseButton )
	self.BossOptions:Show(false)
	bMovable = false
	self.BarContainer:SetStyle("Moveable", false)
	self.BarContainer:SetStyle("Picture", false)
	self.BarContainer:SetStyle("IgnoreMouse", true)
	self.BarContainer:SetText("")
	self.BMPhase:Show(false)
	self.BMTimer[1][2]:Show(false)
	self.BMTimer[2][2]:Show(false)
	self.BMTimer[3][2]:Show(false)
end

function BossMods:OnAlways( wndHandler, wndControl, eMouseButton )
if wndHandler:IsChecked() then
		bmRalways = true
		if not bmNF then
			Apollo.RegisterEventHandler("NextFrame", "OnUpdate", self)
			bmNF = true
		end
	else
		bmRalways = false
		Apollo.RemoveEventHandler("NextFrame", self)
		bmNF = false
	end
end

-----------------------------------------------------------------------------------------------
-- BossMods Instance
-----------------------------------------------------------------------------------------------
local BossModsInst = BossMods:new()
BossModsInst:Init()
